#include <omp.h>
#include <inttypes.h>
#include "mex.h"
#include <simstruc.h> // needed to forward printf output to matlab

/* Computes the global outside density for all given parameter combinations
 * in parallel.
 *
 * This function is intended to be called from Matlab.
 *
 * Start matlab with this LD_PRELOAD
 * LD_PRELOAD=/usr/lib/gcc/x86_64-linux-gnu/4.9.2/libgomp.so matlab
 */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  if (nrhs != 9) {
    mexErrMsgTxt("requires 9 input arguments.");
  } else if (nlhs != 3) {
    mexErrMsgTxt("requires 3 output arguments.");
  }

  double* comm = (double *) mxGetPr(prhs[0]);
  double* partialCoverGraph = (double *) mxGetPr(prhs[6]);
  double* expectedAreaInside =  (double*) mxGetPr(prhs[7]);
  double* observed1sInside =  (double*) mxGetPr(prhs[8]);
  double* Hs = (double*) mxGetPr(prhs[2]);
  double* gammas = (double*) mxGetPr(prhs[3]);
  const int numParamCombis = (int) mxGetNumberOfElements (prhs[3]);
  const int64_t N = (int64_t) mxGetScalar(prhs[1]);
  const double availableArea =  (double) mxGetScalar(prhs[4]);
  const double availableOnes =  (double) mxGetScalar(prhs[5]);

  plhs[0] = mxCreateNumericArray (
          mxGetNumberOfDimensions (prhs[3]), mxGetDimensions(prhs[3]), 
          mxGetClassID(prhs[3]), mxREAL);
  plhs[1] = mxCreateNumericArray (
          mxGetNumberOfDimensions (prhs[3]), mxGetDimensions(prhs[3]), 
          mxGetClassID(prhs[3]), mxREAL);
  plhs[2] = mxCreateNumericArray (
          mxGetNumberOfDimensions (prhs[3]), mxGetDimensions(prhs[3]), 
          mxGetClassID(prhs[3]), mxREAL);

  double* expectedAreaOutside = mxGetPr (plhs[0]);
  double* observed1sOutside = mxGetPr (plhs[1]);
  double* observed0sOutside = mxGetPr (plhs[2]);

#pragma omp parallel for
  for (int paramCombi = 0; paramCombi < numParamCombis; ++paramCombi) {
    const int64_t H = (int64_t) Hs[paramCombi];
    const int64_t gamma = (int64_t) gammas[paramCombi];

    int partialCoverGraphAndModel = 0;
    int commAndPartialCoverGraphAndModel = 0;

    for (int i = 0; i<=H; ++i) {
        for (int j = i; j<N; ++j) {
            const int index = i*N+j;
            partialCoverGraphAndModel += 
                    (int)(partialCoverGraph[index] == 1.);
            commAndPartialCoverGraphAndModel += 
                    (int)((partialCoverGraph[index] == 1.) && (comm[index] == 1.));
        }
    }

    for (int i = H+1; i<=gamma; ++i) {
        const int64_t gamma2NH = gamma*gamma-(N-1)*H;
        const int64_t NH2gamma = N-1+H - 2*gamma;
        const int64_t iNHgamma = i*NH2gamma+gamma2NH;
        const int64_t maxIndex = ((gamma-H)*(gamma-H)*(gamma-(N-1))*(gamma-(N-1))-gamma2NH*iNHgamma) / (NH2gamma*iNHgamma);
        for (int j = i; j<= maxIndex; ++j) {
            const int index = i*N+j;
            partialCoverGraphAndModel += 
                (int)(partialCoverGraph[index] == 1.);
            commAndPartialCoverGraphAndModel += 
                (int)((partialCoverGraph[index] == 1.) && (comm[index] == 1.));
        }     
    }
    
    expectedAreaOutside[paramCombi] = availableArea - expectedAreaInside[paramCombi] + (double) partialCoverGraphAndModel;
    observed1sOutside[paramCombi] = availableOnes - observed1sInside[paramCombi] + (double) commAndPartialCoverGraphAndModel;
    observed0sOutside[paramCombi] = expectedAreaOutside[paramCombi] - observed1sOutside[paramCombi];
    //printf("H, gamma: %i, %i --> %i, %i, %.0f\n", H, gamma, partialCoverGraphAndModel, commAndPartialCoverGraphAndModel, observed1sOutside[paramCombi]);

  }
  
  return;
}
